# MCMC FOR PROBLEMS WITH FAST AND SLOW VARIABLES BY DRAGGING FAST VARIABLES.
# 
# Radford M. Neal, 2010.


# METROPOLIS UPDATES WITH FAST VARIABLE DRAGGING.
# 
# Arguments are as follows:
#
#   x1        initial values for the slow variables 
#   x2        initial values for the fast variables 
#   s1        standard deviations for the normal proposal for x1
#   s2        standard deviations for the normal proposals for x2
#   iters     number of slow variable updates
#   rep       number of fast variable updates during dragging, also the number
#             of interpolating distributions; may be zero (x2 not changed)
#   lpr       function that computes the log probability density for the 
#             joint distribution of x1 and x2
#
#   extra     number of extra Metropolis updates for x2 to do after each
#             dragging update (default 0)
#   update.1  if TRUE (default is FALSE), update elements of x1 (outer) 
#             and x2 (inner) one at a time. The total number of updates 
#             is therefore greater, and separate rejection rates are 
#             maintained for each element.
#
# The s1 and s2 arguments may be scalars (same standard deviation for
# all components) or vectors the same lengths as x1 and x2.
#
# The lpr function for computing the log density must be as documented
# in ens-met.r.
#
# The value returned is a list with matrix elements "X1" and "X2
# containing the iters+1 values from the chain (one per row, including
# the initial value) and elements "rej" containing the rejection
# rate(s) for outer the Metropolis updates of x1, and "rej2" containing 
# rejection rate(s) for the inner updates of x2.

drag_metropolis <- function (x1, x2, s1, s2, iters, rep, lpr, 
                             extra=0, update.1=FALSE)
{
  # Create matrices to store simulation results, and put in initial state.

  X1 <- matrix(0,iters+1,length(x1))
  X1[1,] <- x1
  X2 <- matrix(0,iters+1,length(x2))
  X2[1,] <- x2

  # Set current state to initial state.

  curr.x1 <- x1
  curr.x2 <- x2

  curr.lpr <- lpr (curr.x1, curr.x2, NULL)

  # Initialize for loop.

  if (update.1)
  { rej <- rep(0,length(x1))
    rej2 <- rep(0,length(x2))
    s1 <- rep(s1,length=length(x1))
    s2 <- rep(s2,length=length(x2))
  }
  else
  { rej <- 0
    rej2 <- 0
  }

  # Loop that does iters Metropolis updates with x2 dragged.

  for (i in seq_len(iters))
  { 
    if (update.1)
    {
      # Update one variable of x1 at a time.

      for (e in seq_len(length(x1)))
      {
        # Propose a change to the e'th component of x1.

        prop.x1 <- curr.x1    
        prop.x1[e] <- rnorm (1, prop.x1[e], s1[e])
        prop.lpr <- lpr (prop.x1, curr.x2, NULL)
    
        # Perform a series of dragging updates with the slow part of the dragged
        # state at curr.x1 & prop.x1 and the fast part starting at curr.x2 and
        # ending at prop.x2.  Accumulate the sum, s, needed to compute the 
        # acceptance probability.
    
        drag.x2 <- curr.x2
     
        drag.lpr0 <- curr.lpr
        drag.lpr1 <- prop.lpr
    
        s <- drag.lpr1 - drag.lpr0
    
        for (h in seq(1,length=rep))
        { 
          # Propose new values for components of x2 in turn, and accept or 
          # reject in the usual Metropolis fashion, using the interpolated 
          # log density function.

          for (u in seq_len(length(x2)))
          {
            drag.prop.x2 <- drag.x2
            drag.prop.x2[u] <- rnorm (1, drag.x2[u], s2[u])
      
            drag.prop.lpr0 <- lpr (curr.x1, drag.prop.x2, curr.lpr)
            drag.prop.lpr1 <- lpr (prop.x1, drag.prop.x2, prop.lpr)
      
            a <- (h/(rep+1)) * (drag.prop.lpr1 - drag.lpr1) +
                 (1-h/(rep+1)) * (drag.prop.lpr0 - drag.lpr0)
      
            if (runif(1) < exp(a))
            { drag.x2 <- drag.prop.x2
              drag.lpr0 <- drag.prop.lpr0
              drag.lpr1 <- drag.prop.lpr1
            }
            else
            { rej2[u] <- rej2[u] + 1
            }
          }
    
          s <- s + drag.lpr1 - drag.lpr0
        }
    
        # Decide whether or not to accept the proposed x1 along with the final 
        # dragged value for x2.
    
        prop.x2 <- drag.x2
    
        if (runif(1) < exp(s/(rep+1)))
        { curr.x1 <- prop.x1
          curr.x2 <- prop.x2
          curr.lpr <- lpr (curr.x1, curr.x2, prop.lpr)
        }
        else
        { rej[e] <- rej[e] + 1
        }
      }
    }
    else # not update.1
    {
      # Propose a change to all of x1 from its current value.
  
      prop.x1 <- rnorm (length(x1), curr.x1, s1)
      prop.lpr <- lpr (prop.x1, curr.x2, NULL)
  
      # Perform a series of dragging updates with the slow part of the dragged 
      # state at curr.x1 & prop.x1 and the fast part starting at curr.x2 and
      # ending at prop.x2.  Accumulate the sum, s, needed to compute the 
      # acceptance probability.
  
      drag.x2 <- curr.x2
   
      drag.lpr0 <- curr.lpr
      drag.lpr1 <- prop.lpr
  
      s <- drag.lpr1 - drag.lpr0
  
      for (h in seq(1,length=rep))
      { 
        # Propose a new value for x2, and accept or reject in the usual
        # Metropolis fashion, using the interpolated log density function.
  
        drag.prop.x2 <- rnorm (length(x2), drag.x2, s2)
  
        drag.prop.lpr0 <- lpr (curr.x1, drag.prop.x2, curr.lpr)
        drag.prop.lpr1 <- lpr (prop.x1, drag.prop.x2, prop.lpr)
  
        a <- (h/(rep+1)) * (drag.prop.lpr1 - drag.lpr1) +
             (1-h/(rep+1)) * (drag.prop.lpr0 - drag.lpr0)
  
        if (runif(1) < exp(a))
        { drag.x2 <- drag.prop.x2
          drag.lpr0 <- drag.prop.lpr0
          drag.lpr1 <- drag.prop.lpr1
        }
        else
        { rej2 <- rej2 + 1
        }
  
        s <- s + drag.lpr1 - drag.lpr0
      }
  
      # Decide whether or not to accept the proposed x1 along with the final 
      # dragged value for x2.
  
      prop.x2 <- drag.x2
  
      if (runif(1) < exp(s/(rep+1)))
      { curr.x1 <- prop.x1
        curr.x2 <- prop.x2
        curr.lpr <- lpr (curr.x1, curr.x2, prop.lpr)
      }
      else
      { rej <- rej + 1
      }
    }
  
    # Do extra Metropolis updates on the single state, if asked to.

    if (extra>0)
    {
      for (j in seq_len(extra))
      { 
        if (update.1)
        {
          # Update x2 one variable at a time.

          for (e in seq_len(length(x2)))
          { 
            prop.x2 <- curr.x2
            prop.x2[e] <- rnorm(1,prop.x2[e],s2[e])

            prop.lpr <- lpr(curr.x1,prop.x2,curr.lpr)

            if (runif(1) < exp (prop.lpr-curr.lpr))
            { curr.x2 <- prop.x2
              curr.lpr <- prop.lpr
            }
            else
            { rej2[e] <- rej2[e] + 1
            }
          }
        }
        else # not update.1
        {
          # Update all of x2 simultaneously.

          prop.x2 <- rnorm(length(x2),curr.x2,s2)

          prop.lpr <- lpr(curr.x1,prop.x2,curr.lpr)
  
          if (runif(1) < exp (prop.lpr-curr.lpr))
          { curr.x2 <- prop.x2
            curr.lpr <- prop.lpr
          }
          else
          { rej2 <- rej2 + 1
          }
        }
      }
    }

    X1[i+1,] <- curr.x1
    X2[i+1,] <- curr.x2
  }
  
  # Return the state at each iteration and the average rejection rates.

  if (update.1)
  { list (X1=X1, X2=X2, rej=rej/iters, rej2=rej2/(iters*(rep+extra)*length(x1)))
  }
  else
  { list (X1=X1, X2=X2, rej=rej/iters, rej2=rej2/(iters*(rep+extra)))
  }
}
