# ENSEMBLE METROPOLIS FOR PROBLEMS WITH FAST AND SLOW VARIABLES.
#
# Radford M. Neal, 2010.
#
# This file contains the function for performing ensemble MCMC with
# fast and slow variables.  Three ensemble base distributions for use
# with it are also defiend here.  The user may define other ensemble
# base distributions, and must define the log density function
# defining the distribution to sample from.


# LOG DENSITY FUNCTION FOR ENSEMBLE MCMC.  To use ensemble MCMC, the
# user must provide a fucntion that computes the log probability
# density for the distribution.  This function takes a vector value
# for the slow variables, x1, and a vector value for fast variables,
# x2, or a matrix of values for x2 (one per row), and returns the log
# probability or probability density (plus any constant) for (x1,x2),
# or a vector of these log densities for all the values of x2.
#
# Preferrably, the value returned should have attributes that save
# intermediate results of this computation.  When the ensemble MCMC
# function calls the log density function for the same value of x1 as
# previously, it passes the value returned then (with its attributes)
# as the third argument.  This allows this subsequent call to be done
# more quickly.  The third argument will be NULL if the result of a
# previous call with the same value for x1 is not available.


# ENSEMBLE BASE DISTRIBUTIONS.  An ensemble base distribution is
# specified by giving a function that takes a vector of values for the
# fast variables (x2) and the size of the ensemble (K) as arguments,
# and returns a matrix with $K$ rows containing values for the fast
# variables in the ensemble.  A randomly chosen row in this matrix
# will contain the current value (x2), with other rows drawn from the
# conditional distribution given that the current value is in the
# ensemble.  The function may take extra arguments that are parameters
# of the ensemble base distribution.  The matrix returned may have an
# attribute "lp", which is a vector of length K giving the logs of the
# marginal ensemble base densities for the ensemble members (times any
# constant).  If this attribute is not present, these densities are
# assumed to be the same (and may possibly be improper).
#
# Functions for three ensemble base distributions are supplied below.


# ENSEMBLE OF INDEPENDENT NORMALS.  With means and standard deviations
# given by mu and sd (scalars or vectors the same length as x2).

ens_cond_ind <- function (x2, K, mu, sd)
{
  ens <- matrix(NA,K,length(x2))
  lp <- rep(NA,K)
  k <- sample(seq_len(K),1)

  for (i in seq_len(K))
  { if (i==k) 
    { ens[i,] <- x2
    }
    else
    { ens[i,] <- rnorm(length(x2),mu,sd)
    }
    lp[i] <- sum(dnorm(ens[i,],mu,sd,log=TRUE))
  }

  attr(ens,"lp") <- lp

  ens
}


# ENSEMBLE OF IMPROPER EXCHANGEABLE VALUES.  From normal distribution
# with unknown mean (improper uniform prior) and standard distribution
# sd (a scalar or vector of the same length as x2).

ens_cond_exch <- function (x2, K, sd)
{
  ens <- matrix(NA,K,length(x2))
  k <- sample(seq_len(K),1)

  mu <- rnorm(length(x2),x2,sd)

  for (i in seq_len(K))
  { if (i==k) 
    { ens[i,] <- x2
    }
    else
    { ens[i,] <- rnorm(length(x2),mu,sd)
    }
  }

  ens
}


# ENSEMBLE OF VALUES ON A GRID.  The grid is defined by a function that
# returns a matrix with number of rows equal to K, each row giving an offset 
# for x2.  The function ens.rect.grid defined below is suitable.

ens_cond_grid <- function (x2, K, grid)
{
  gr <- grid()

  if (nrow(gr)!=K)
  { stop("Grid must have number of rows equal to size of ensemble")
  }

  ens <- matrix(NA,K,length(x2))
  k <- sample(seq_len(K),1)

  for (i in seq_len(K))
  { ens[i,] <- x2 + gr[i,] - gr[k,]
  }

  ens
}


# MAKE RECTANGULAR GRID.  This function returns a grid, and is suitable 
# for ens.cond.grid.  Arguments are a vector with the number of grid
# points for each fast variable, a vector with the low value for the
# grid extent for each fast variable (or a scalar applying to all),
# and a vector with the high value for the grid extent (default the
# same as the low values).  The actual grid extent for each fast
# variable is chosen uniformly from the low to the high values.  The
# grid spacing for a variable is the grid extent divided by the number
# of grid points minus 1.  The number of grid points must be at least
# two, unless the high value for the grid extent for that variable is
# zero, in which case one grid point is allowed.

rect_grid <- function (n, low, high=low)
{
  if (length(low)!=1 && length(low)!=length(n) 
   || length(high)!=1 && length(high)!=length(n) 
   || any(high<0) || any(low>high) 
   || any(n<1) || any(n<2 & high>0))
  { stop("Invalid argument to rect_grid")
  }

  extent <- runif(length(n),low,high)
  spacing <- extent / (n-1)
  spacing[n==1] <- 0

  K <- prod(n)
  grid <- matrix(NA,K,length(n))

  for (i in seq_len(K))
  { x <- i-1
    for (j in seq_len(length(n)))
    { grid[i,j] <- spacing[j] * x%%n[j]
      x <- x %/% n[j]
    }
  }

  grid
}


# ENSEMBLE MCMC USING METROPOLIS PROPOSALS.  Does updates in which a
# single state (x1, x2) is mapped to an ensemble with multiple x2
# values, Metropolis updates are done on the ensemble, and the
# ensemble is then mapped backto a single state.  The x1 variables are
# "slow", with the x2 variables being "fast".
#
# Arguments are as follows:
#
#   x1        initial values for the slow variables 
#   x2        initial values for the fast variables 
#   s1        standard deviations for the normal proposal for x1
#   s2        standard deviations for the shift in the x2 ensemble (may be 0)
#   iters     number of iterations of map to ensemble / update / map back
#   rep       number of updates to do on the ensemble before mapping back
#   lpr       function that computes the log probability density for the 
#             joint distribution of x1 and x2
#   K         size of the ensemble
#   ens.cond  conditional distribution for ensemble given current state
#   ...       further arguments that are passed to ens.cond
#
#   update.1  if TRUE (default is FALSE), update elements of x1 one at a 
#             time, each with a corresponding update of the ensemble.
#             The total number of updates is therefore greater, and 
#             separate rejection rates are maintained for each element.
#
#   extra     number (default 0) of extra Metropolis updates on fast variables 
#             to do on the single state after mapping back from the ensemble
#   xs2       standard deviations for the normal proposal for x2 in these
#             extra updates (default is s2)
#
# The s1 and s2 arguments may be scalars (same standard deviation for
# all components) or vectors the same lengths as x1 and x2.  If s2 is
# not all zero, the value returned by the ens.cond function must not
# have a "pr" attribute - ie, the ensemble base measure must be
# invariant under a shift of all ensemble members by the same amount.
#
# The lpr function for computing the log density must be as documented
# at the top of this file.
#
# The value returned is a list with matrix elements "X1" and "X2
# containing the iters+1 values from the chain (one per row, including
# the initial value) and component "rej" containing the rejection rate(s)
# for the Metropolis updates.

ens_metropolis <- function (x1, x2, s1, s2, iters, rep, lpr, K, ens_cond, ...,
                            update.1=FALSE, extra=0, xs2=s2)
{
  # Create matrices to store simulation results, and put in initial state.

  X1 <- matrix(0,iters+1,length(x1))
  X1[1,] <- x1
  X2 <- matrix(0,iters+1,length(x2))
  X2[1,] <- x2

  # Set current state to initial state.

  curr.x1 <- x1
  curr.x2 <- x2

  curr.lpr <- NULL

  # Define function to evaluate the log probability of an ensemble.  Takes
  # slow variables (x1), matrix of fast variables (ens), vector of ensemble
  # base measure probabilities, and result of a previous lpr evaluation with 
  # these slow variables (or NULL) as arguments.  Returns a list with the
  # relative proabilities for ensemble members (P), the log of the ensemble
  # probability (log.sum.P), and the result of the lpr call as components.

  ens.lpr <- function (x1, ens, elp, prev.lpr)
  { ens.lpr <- lpr(x1,ens,prev.lpr)
    if (!is.null(elp))
    { ens.lpr <- ens.lpr - elp
    }
    max.lpr <- max(ens.lpr)
    P <- exp(ens.lpr-max.lpr)
    log.sum.P <- log(sum(P)) + max.lpr
    list (P=P, log.sum.P=log.sum.P, lpr=ens.lpr)
  }

  # Initialize for loop.

  if (update.1)
  { rej <- rep(0,length(x1))
    s1 <- rep(s1,length=length(x1))
    s2 <- rep(s2,length=length(x2))
    if (extra>0) 
    { rej2 <- rep(0,length(x2))
      xs2 <- rep(xs2,length=length(x2))
    }
  }
  else
  { rej <- 0
    if (extra>0)
    { rej2 <- 0
    }
  }

  # Loop that does iters ensemble updates, regenerating the ensemble for each.

  for (i in seq_len(iters))
  { 
    # Generate an ensemble of x2 values, and find its ensemble probability.

    curr.ens <- ens_cond (curr.x2, K, ...)
    ens.elp <- attr(curr.ens,"lp")
    if (!is.null(ens.elp) && any(s2!=0))
    { stop("Ensemble distribution must be shift invariant if s2 is not zero")
    }

    curr.lpr <- ens.lpr (curr.x1, curr.ens, ens.elp, curr.lpr$lpr)

    # Do rep Metropolis updates for the ensemble.

    for (j in seq_len(rep))
    { 
      if (update.1)
      {
        for (e in seq_len(length(x1)))
        { 
          # Update x1 one variable at a time, proposing a new value for 
          # e'th component of x1 and the ensemble of x2 values (same as old
          # if s2 is all zeros).
    
          prop.x1 <- curr.x1
          prop.x1[e] <- rnorm(1,prop.x1[e],s1)

          if (any(s2!=0))
          { prop.ens <- curr.ens + rep (rnorm(length(x2),0,s2), each=K)
          }
          else
          { prop.ens <- curr.ens
          }
    
          # Find the ensemble probability for the proposal.
    
          prop.lpr <- ens.lpr (prop.x1, prop.ens, ens.elp, NULL)
    
          # Accept or reject the proposal.
    
          if (runif(1) < exp (prop.lpr$log.sum.P-curr.lpr$log.sum.P))
          { curr.x1 <- prop.x1
            curr.ens <- prop.ens
            curr.lpr <- prop.lpr
          }
          else
          { rej[e] <- rej[e] + 1
          }
        }
      }
      else # not update.1
      {
        # Update x1 all at once, proposing new values for x1 and the ensemble 
        # of x2 values (same as old if s2 is all zeros).
  
        prop.x1 <- rnorm(length(x1),curr.x1,s1)
        if (any(s2!=0))
        { prop.ens <- curr.ens + rep (rnorm(length(x2),0,s2), each=K)
        }
        else
        { prop.ens <- curr.ens
        }
  
        # Find the ensemble probability for the proposal.
  
        prop.lpr <- ens.lpr (prop.x1, prop.ens, ens.elp, NULL)
  
        # Accept or reject the proposal.
  
        if (runif(1) < exp (prop.lpr$log.sum.P-curr.lpr$log.sum.P))
        { curr.x1 <- prop.x1
          curr.ens <- prop.ens
          curr.lpr <- prop.lpr
        }
        else
        { rej <- rej+1
        }
      }
    }

    # Select a single state from the ensemble, and store as the current state.

    curr.x2 <- curr.ens[sample(K,1,prob=curr.lpr$P),]

    # Do extra Metropolis updates on the single state, if asked to.

    if (extra>0)
    {
      currx.lpr <- lpr(curr.x1,curr.x2,curr.lpr$lpr)

      for (j in seq_len(extra))
      { 
        if (update.1)
        {
          # Update x2 one variable at a time.

          for (e in seq_len(length(x2)))
          { 
            prop.x2 <- curr.x2
            prop.x2[e] <- rnorm(1,prop.x2[e],xs2[e])

            prop.lpr <- lpr(curr.x1,prop.x2,currx.lpr)

            if (runif(1) < exp (prop.lpr-currx.lpr))
            { curr.x2 <- prop.x2
              currx.lpr <- prop.lpr
            }
            else
            { rej2[e] <- rej2[e] + 1
            }
          }
        }
        else # not update.1
        {
          # Update all of x2 simultaneously.

          prop.x2 <- rnorm(length(x2),curr.x2,xs2)

          prop.lpr <- lpr(curr.x1,prop.x2,currx.lpr)
  
          if (runif(1) < exp (prop.lpr-currx.lpr))
          { curr.x2 <- prop.x2
            currx.lpr <- prop.lpr
          }
          else
          { rej2 <- rej2 + 1
          }
        }
      }
    }

    X1[i+1,] <- curr.x1
    X2[i+1,] <- curr.x2
  }

  # Return state at each iteration and the average rejection rate(s).

  if (extra>0)
  { list (X1=X1, X2=X2, rej=rej/(iters*rep), rej2=rej2/(iters*extra))
  }
  else
  { list (X1=X1, X2=X2, rej=rej/(iters*rep))
  }
}
