# TESTS OF FAST/SLOW METHODS FOR GAUSSIAN PROCESS MODELS USING EIGENVECTORS.
#
# Radford M. Neal, 2010.

source("gp.r")
source("gp-example.r")
source("gp-plot.r")
source("std-met.r")
source("grid-met.r")
source("ens-met.r")
source("drag-met.r")


# RUN TESTS WITH EIGENVECTOR COMPUTATIONS.  Uses the example in gp-example.r.

gp_rune <- function (iters, s1, s2, K, seed=1,
  initial.x1 = ex.nu.mean,
  initial.x2 = c(ex.sigma.mean, ex.eta.mean),
  prefix = "",
  extra = 0,
  which = c(
    "std.met", "std.met1",
    "extra.met", "extra.met1",
    "grid.met", 
    "ens.ind", "ens.ind1",
    "ens.exch", "ens.exch1",
    "ens.grid", "ens.grid1",
    "drag.met", "drag.met1"
))
{
  n1 <- ncol(z)
  n2 <- 2

  cat("\nTESTS ON GP MODEL, EIGENVECTOR COMPUTATIONS\n\n")
  cat ("seed =",seed,
       " n1 =",n1," n2 =",n2,
       " s1 =",s1," s2 =",s2,
       " K =",K,
       " extra =",extra,
       "\n\n")

  dthin <- ceiling(iters/2000)
  
  start.time <- 0

  start <- function (t) 
  { cat("\n",t,"\n\n",sep="")
    lpr.evals <<- 0
    lpr.evals.slow <<- 0
    set.seed(seed)
    start.time <<- proc.time()[1]
  }

  gp_test_eig <- function (z, y, const, jitter, 
    eta.mean, eta.sd, sigma.mean, sigma.sd, nu.mean, nu.sd, nu.sd2,
    s1, s2, iters, rep, method, ...)
  {
    method ( 
      x1 = initial.x1, x2 = initial.x2,
      s1 = s1, s2 = s2, 
      iters = iters, rep = rep,
      lpr = function (x1,x2,prev.lpr) 
        gp_lpr_eig (x1, x2, prev.lpr, z, y, const, jitter, 
          eta.mean, eta.sd, sigma.mean, sigma.sd, nu.mean, nu.sd, nu.sd2),
      ...
    )
  }

  finish <- function (file,thin=dthin)
  { end.time <- proc.time()[1]
    cat("Evaluations: ",lpr.evals.slow,"slow,",lpr.evals,"total\n")
    if (is.null(r$rej2) || !is.finite(r$rej2))
    { cat("Rejection %: ",round(100*r$rej),"\n")
    }
    else
    { cat("Rejection %:",round(100*r$rej),"/",round(100*r$rej2),"\n")
    }
    cat("Time: ",round(end.time-start.time,2),"\n")
    save.image()
    postscript(paste(prefix,file,sep=""),horizontal=TRUE,pointsize=10)
    gp_plot(r,thin=thin)
    dev.off()
    cat("Done\n\n")
  }

  if ("std.met" %in% which)
  { start("Standard Metropolis")
    r.std.met <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, s2, iters, 1, std_metropolis)
    finish("std-met.ps")
  }

  if ("std.met1" %in% which)
  { start("Standard Metropolis, one-variable updates")
    r.std.met1 <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, s2, ceiling(iters/length(initial.x1)), 1, 
                   std_metropolis, update.1=TRUE)
    finish("std-met1.ps",thin=ceiling(dthin/length(initial.x1)))
  }

  if ("extra.met" %in% which)
  { start("Standard Metropolis with extra fast updates")
    r.extra.met <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, s2, iters, K+extra, std_metropolis)
    finish("extra-met.ps")
  }

  if ("extra.met1" %in% which)
  { start("Standard Metropolis with extra fast updates, one-variable updates")
    r.extra.met1 <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, s2, ceiling(iters/length(initial.x1)), K+extra, 
                   std_metropolis, update.1=TRUE)
    finish("extra-met1.ps",thin=ceiling(dthin/length(initial.x1)))
  }

  if ("grid.met" %in% which)
  { start("Grid Metropolis")
    r.grid.met <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, s2, ceiling(iters/5), K, grid_metropolis)
    finish("grid-met.ps")
  }

  if ("ens.ind" %in% which)
  { start("Ensemble Metropolis, independent ensemble")
    r.ens.ind <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, 0, iters, 1, 
                   ens_metropolis, K,
                     ens_cond_ind, mu=c(ex.sigma.mean,ex.eta.mean), 
                                   sd=c(ex.sigma.sd,ex.eta.sd), 
                     extra=extra, xs2=s2)
    finish("ens-ind.ps")
  }

  if ("ens.ind1" %in% which)
  { start("Ensemble Metropolis, independent ensemble, one-variable updates")
    r.ens.ind1 <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, 0, ceiling(iters/length(initial.x1)), 1, 
                   ens_metropolis, K, 
                     ens_cond_ind, mu=c(ex.sigma.mean,ex.eta.mean), 
                                   sd=c(ex.sigma.sd,ex.eta.sd),
                     extra=extra, xs2=s2,
                     update.1=TRUE)
    finish("ens-ind1.ps",thin=ceiling(dthin/length(initial.x1)))
  }

  if ("ens.exch" %in% which)
  { start("Ensemble Metropolis, exchangeable ensemble")
    r.ens.exch <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, 0, iters, 1, 
                   ens_metropolis, K,
                     ens_cond_exch, sd=0.4*c(ex.sigma.sd,ex.eta.sd),
                     extra=extra, xs2=s2)
    finish("ens-exch.ps")
  }

  if ("ens.exch1" %in% which)
  { start("Ensemble Metropolis, exchangeable ensemble, one-variable updates")
    r.ens.exch1 <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, 0, ceiling(iters/length(initial.x1)), 1, 
                   ens_metropolis, K,
                     ens_cond_exch, sd=0.4*c(ex.sigma.sd,ex.eta.sd),
                     extra=extra, xs2=s2,
                     update.1=TRUE)
    finish("ens-exch1.ps",thin=ceiling(dthin/length(initial.x1)))
  }

  if ("ens.grid" %in% which)
  { start("Ensemble Metropolis, grid ensemble")
    r.ens.grid <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, 0, iters, 1, 
                   ens_metropolis, K,
                    ens_cond_grid, 
                    grid = function () rect_grid (rep(round(sqrt(K)),2), 
                    1.0*c(ex.sigma.sd,ex.eta.sd), 1.1*c(ex.sigma.sd,ex.eta.sd)),
                    extra=extra, xs2=s2)
    finish("ens-grid.ps")
  }

  if ("ens.grid1" %in% which)
  { start("Ensemble Metropolis, grid ensemble, one-variable updates")
    r.ens.grid1 <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, 0, ceiling(iters/length(initial.x1)), 1, 
                   ens_metropolis, K,
                    ens_cond_grid, 
                     grid = function () rect_grid (rep(round(sqrt(K)),2),
                    1.0*c(ex.sigma.sd,ex.eta.sd), 1.1*c(ex.sigma.sd,ex.eta.sd)),
                    extra=extra, xs2=s2,
                    update.1=TRUE)
    finish("ens-grid1.ps",thin=ceiling(dthin/length(initial.x1)))
  }

  if ("drag.met" %in% which)
  { start("Dragging Metropolis")
    r.drag.met <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, s2, iters, K, drag_metropolis, extra=extra)
    finish("drag-met.ps")
  }

  if ("drag.met1" %in% which)
  { start("Dragging Metropolis, one-variable updates")
    r.drag.met1 <<- r <- 
      gp_test_eig (z, y, ex.const, ex.jitter, 
                   ex.eta.mean, ex.eta.sd, ex.sigma.mean, ex.sigma.sd,
                   ex.nu.mean, ex.nu.sd, ex.nu.sd2,
                   s1, s2, ceiling(iters/length(initial.x1)), K, 
                   drag_metropolis, extra=extra, update.1=TRUE)
    finish("drag-met1.ps",thin=ceiling(dthin/length(initial.x1)))
  }
}
