# LOG DENSITY FUNCTION FOR GP REGRESSION MODELS.
#
# Radford M. Neal, 2010.

lpr.evals <- 0
lpr.evals.slow <- 0


# GAUSSIAN PROCESS REGRESSION MODEL.  The covariance between response i and
# response j, with covariates zi and zj, is
#
#   eta^2 * (const^2 + exp (- SUM_k (nu_k*(zi_k-zj_k))^2) + delta(i,j)*jitter^2)
#    + delta(i,j) sigma^2
#
# Here, delta(i,j) is one if i=j and zero otherwise, const and jitter have
# specified fixed values, and eta, sigma, and nu are hyperparameters.  The
# priors for eta and sigma are as follows:
# 
#   log(eta)   ~  N(eta.mean,eta.sd^2)
#   log(sigma) ~  N(sigma.mean,sigma.sd^2)
#
# The relevances, nu, are the same for all k if nu.mean is scalar, with
#
#   log(nu)    ~  N(nu.mean,nu.sd)
#
# Otherwise, each nu_k is different, with 
#
#   log(nu)    ~  MVN(nu.mean,nu.Sigma)
#
# with u.Sigma having nu.sd^2 + nu.sd2^2 on its diagonal and nu.sd2^2 in 
# all off-diagonal entries.


# COMPUTE CORE PART OF COVARIANCE MATRIX.  Leaves out the scaling by eta^2
# and the addition of noise of standard deviation sigma.

gp_core_cov0 <- function (z, const, jitter, nu)
{
  n <- nrow(z)
  M <- matrix(0,n,n)

  csq <- const^2
  jsq <- jitter^2

  for (i in seq_len(n))
  { zi <- z[i,]
    for (j in i:n)
    { M[i,j] <- M[j,i] <- csq + exp(-sum((nu*(zi-z[j,]))^2))
    }
    M[i,i] <- M[i,i] + jsq
  }

  M
}

gp_core_cov1 <- function (z, const, jitter, nu)
{
  n <- nrow(z)
  M <- matrix(0,n,n)
  tz <- t(z)

  csq <- const^2

  for (i in seq_len(n))
  { M[,i] <- csq + exp(-colSums((nu*(tz-tz[,i]))^2))
  }

  diag(M) <- diag(M) + jitter^2

  M
}

gp_core_cov <- function (z, const, jitter, nu)
{
  n <- nrow(z)
  M <- matrix(0,n,n)

  nu.tz <- nu * t(z)
  csq <- const^2

  for (i in seq_len(n))
  { D <- nu.tz - nu.tz[,i]
    M[,i] <- csq + exp(-colSums(D*D))
  }

  diag(M) <- diag(M) + jitter^2

  M
}


# LOG OF PRIOR DENSITY FOR RELEVANCE HYPERPARAMETERS.  May omit constant term.

gp_nu_prior <- function (x, nu.mean, nu.sd, nu.sd2)
{ 
  if (length(nu.mean)!=1 && length(nu.mean)!=length(x)
   || length(nu.sd)!=1 || length(nu.sd2)!=1)
  { stop("Invalid mean/sd/sd2 for nu")
  }

  if (nu.sd2==0)
  { nu.lpd <- sum(dnorm(x,nu.mean,nu.sd,log=TRUE))
  }
  else
  { p <- length(x)
    M <- matrix ( (1/(nu.sd^2+n*nu.sd2^2) - 1/nu.sd^2) / p, p, p)
    diag(M) <- diag(M) + 1/nu.sd^2
    x <- x - nu.mean
    nu.lpd <- - t(x) %*% M %*% x / 2
  }

  as.vector (nu.lpd)
}


# UPDATE GP REGRESSION MODEL USING EIGENVECTORS.  The slow variables (x1) 
# are log(nu) or the log(nu_k).  The fast variables (x2) are log(sigma) 
# and log(eta).  The data is in the matrix z (covariates) and vector y 
# (responses).  Note that x1 is a vector giving a single set of values
# for slow parameters, but x2 may be a matrix, with each row giving a set
# of values for the fast parameters.

gp_lpr_eig <- function (x1, x2, prev.lpr, z, y,
                        const, jitter,
                        eta.mean, eta.sd,
                        sigma.mean, sigma.sd,
                        nu.mean, nu.sd, nu.sd2=0)
{
  if (!is.matrix(x2)) 
  { x2 <- matrix(x2,1,length(x2))
  }

  n <- nrow(z)
  p <- ncol(z)

  if (length(y)!=n)
  { stop("Mismatch of z and y")
  }
  
  if (length(x1)!=1 && length(x1)!=p || ncol(x2)!=2)
  { stop("Invalid parameter dimensions")
  }

  lpr.evals <<- lpr.evals + nrow(x2)

  e.values <- attr(prev.lpr,"e.values")
  proj.sq <- attr(prev.lpr,"proj.sq")
  nu.lpd <- attr(prev.lpr,"nu.lpd")

  if (is.null(e.values) || is.null(proj.sq) || is.null(nu.lpd))
  { 
    lpr.evals.slow <<- lpr.evals.slow + 1

    nu <- exp(x1)
    nu.lpd <- gp_nu_prior (x1, nu.mean, nu.sd, nu.sd2)

    M <- gp_core_cov (z, const, jitter, nu)
    eig <- eigen (M, symmetric=TRUE)
    e.values <- eig$values
    proj.sq <- as.vector(y %*% eig$vectors)^2
  }
 
  V <- t (exp(2*x2[,1]) + exp(2*x2[,2]) *
            matrix (e.values, nrow(x2), length(e.values), byrow=TRUE))

  lpr <- - colSums (log(V) + proj.sq/V) / 2 + nu.lpd +
         dnorm(x2[,1],sigma.mean,sigma.sd,log=TRUE) +
         dnorm(x2[,2],eta.mean,eta.sd,log=TRUE)

  attr(lpr,"e.values") <- e.values
  attr(lpr,"proj.sq") <- proj.sq
  attr(lpr,"nu.lpd") <- nu.lpd

  lpr
}


# EXTRACT PARAMETERS FROM STATE OR STATES OF EIGENVECTOR METHOD.

gp_eig_extract <- function (x1, x2)
{ if (is.matrix(x2))
  { if (ncol(x2)!=2)
    { stop("Bad states")
    }
    list (nu=exp(x1), eta=exp(x2[,2,drop=FALSE]), sigma=exp(x2[,1,drop=FALSE]))
  }
  else
  { if (length(x2)!=2)
    { stop("Bad state")
    }
    list (nu=exp(x1), eta=exp(x2[2]), sigma=exp(x2[1]))
  }
}


# UPDATE GP REGRESSION MODEL USING CHOLESKY DECOMPOSITION.  The slow variables 
# (x1) are log(nu) or the log(nu_k) and log(sigma)-log(eta).  The fast variable
# (x2), is log(eta).  The data is in the matrix z (covariates) and vector 
# y (responses).  Note that x1 is a vector giving a single set of values
# for slow parameters, but x2 may be a matrix (with one column), with each row 
# giving a value for the fast parameter.

gp_lpr_chol <- function (x1, x2, prev.lpr, z, y,
                        const, jitter,
                        eta.mean, eta.sd,
                        sigma.mean, sigma.sd,
                        nu.mean, nu.sd, nu.sd2=0)
{
  if (!is.matrix(x2)) 
  { x2 <- matrix(x2,1,length(x2))
  }

  n <- nrow(z)
  p <- ncol(z)

  if (length(y)!=n)
  { stop("Mismatch of z and y")
  }
  
  if (length(x1)!=2 && length(x1)!=p+1 || ncol(x2)!=1)
  { stop("Invalid parameter dimensions")
  }

  lpr.evals <<- lpr.evals + nrow(x2)

  log.det <- attr(prev.lpr,"log.det")
  sum.sq <- attr(prev.lpr,"sum.sq")
  nu.lpd <- attr(prev.lpr,"nu.lpd")

  if (is.null(log.det) || is.null(sum.sq) || is.null(nu.lpd))
  { 
    lpr.evals.slow <<- lpr.evals.slow + 1

    nu <- exp(x1[-length(x1)])
    nu.lpd <- gp_nu_prior (x1[-length(x1)], nu.mean, nu.sd, nu.sd2)

    M <- gp_core_cov (z, const, jitter, nu)
    diag(M) <- diag(M) + exp (2*x1[length(x1)])

    ch <- chol (M)
    w <- forwardsolve(ch,y,upper.tri=TRUE,transpose=TRUE)

    log.det <- sum(log(diag(ch)))
    sum.sq <- sum(w^2)

  }

  eta <- exp(x2[,1])

  lpr <- - log.det - n*log(eta) - sum.sq/(2*eta^2) + nu.lpd +
         dnorm(x1[length(x1)]+x2[,1],sigma.mean,sigma.sd,log=TRUE) +
         dnorm(x2[,1],eta.mean,eta.sd,log=TRUE)

  attr(lpr,"log.det") <- log.det
  attr(lpr,"sum.sq") <- sum.sq
  attr(lpr,"nu.lpd") <- nu.lpd

  lpr
}


# EXTRACT PARAMETERS FROM STATE OR STATES OF CHOLESKY METHOD.

gp_chol_extract <- function (x1, x2)
{ if (is.matrix(x2))
  { if (ncol(x2)!=1)
    { stop("Bad states")
    }
    list (nu=exp(x1[,-ncol(x1),drop=FALSE]), eta=exp(x2[,1,drop=FALSE]), 
          sigma=exp(x1[,ncol(x1),drop=FALSE]+x2[,1,drop=FALSE]))
  }
  else
  { if (length(x2)!=1)
    { stop("Bad state")
    }
    list (nu=exp(x1[-length(x1)]), eta=exp(x2[1]), 
          sigma=exp(x1[length(x1)]+x2[1]))
  }
}
