# THE GRID METROPOLIS METHOD FOR SAMPLING WITH FAST AND SLOW VARIABLES.
#
# Radford M. Neal, 2010.


# GRID METROPOLIS METHOD. Samples for (x1, x2) using proposals that
# change x1 and x2 simultaneously, but in groups, in which x1 changes
# in one group are on a grid with a random orientation and with random
# spacing, so that "slow" computations will often be avoided.
#
# Arguments are as follows:
#
#   x1        initial values for the slow variables 
#   x2        initial values for the fast variables 
#   s1        standard deviations for the normal proposals for x1
#   s2        standard deviations for the normal proposals for x2
#   iters     the number of groups of Metropolis updates to do
#   rep       the number of updates in each group 
#   lpr       function that computes the log probability density for the 
#             joint distribution of x1 and x2
#
# The s1 and s2 arguements may be scalars (same standard deviation for
# all components) or vectors the same lengths as x1 and x2.
#
# The lpr function takes a vector value for the slow variables, x1,
# and a vector value for the fast variables, x2 and returns the log
# probability density (plus any constant) for (x1,x2).  Preferrably,
# the value returned should have attributes that save intermediate
# results of this computation.  When the grid Metropolis function
# calls the log density function for the same value of x1 as
# previously, it passes the value returned then (with its attributes)
# as the third argument.  This allows this subsequent call to be done
# more quickly.  The third argument will be NULL if the result of a
# previous call with the same value for x1 is not available.
#
# Once at the beginning of each group of updates, a change for x1 is
# randomly sampled from the normal distribution with standard
# deviations given by s1 (with changes for different elements of x1
# being independent).  For later proposals in the group, only the sign
# of this change is randomly selected.
#
# The value returned is a list with components x1 and x2, containing
# the iters+1 values from the chain (including the initial value),
# component rej containing the rejection rate for the Metropolis
# updates, and component evals containing the average number of "slow"
# evaluations per group of updates.

grid_metropolis <- function (x1, x2, s1, s2, iters, rep, lpr)
{
  # Create matrices to store simulation results, and put in initial state.

  X1 <- matrix(0,iters+1,length(x1))
  X1[1,] <- x1
  X2 <- matrix(0,iters+1,length(x2))
  X2[1,] <- x2

  # Set current state to initial state.

  curr.x1 <- x1
  curr.x2 <- x2

  curr.lpr <- lpr(curr.x1,curr.x2,NULL)

  # Loop that does iters grid Metropolis updates.

  rej <- 0
  evals <- 1  # For the first evaluation, since initial curr.lpr is NULL

  for (i in seq_len(iters))
  { 
    old.x1 <- curr.x1
    old.lpr <- curr.lpr

    dx1 <- rnorm(length(x1),0,s1)

    curr.i <- 0
    min.i <- 0
    max.i <- 0

    pos.lpr <- list()
    neg.lpr <- list()

    for (j in seq_len(rep))
    { 
      prop.i <- curr.i + sample(c(-1,+1),1)
      prop.x1 <- old.x1 + prop.i*dx1
      prop.x2 <- rnorm(length(x2),curr.x2,s2)

      if (prop.i<min.i) 
      { min.i <- prop.i
        prop.lpr <- lpr(prop.x1,prop.x2,NULL)
        neg.lpr[[-min.i]] <- prop.lpr
        evals <- evals+1
      }
      else if (prop.i>max.i) 
      { max.i <- prop.i
        prop.lpr <- lpr(prop.x1,prop.x2,NULL)
        pos.lpr[[max.i]] <- prop.lpr
        evals <- evals+1
      }
      else if (prop.i<0)
      { prop.lpr <- lpr(prop.x1,prop.x2,neg.lpr[[-prop.i]])
      }
      else if (prop.i>0)
      { prop.lpr <- lpr(prop.x1,prop.x2,pos.lpr[[prop.i]])
      }
      else 
      { prop.lpr <- lpr(prop.x1,prop.x2,old.lpr)
      }

      if (runif(1) < exp (prop.lpr-curr.lpr))
      { curr.i <- prop.i
        curr.x1 <- prop.x1
        curr.x2 <- prop.x2
        curr.lpr <- prop.lpr
      }
      else
      { rej <- rej+1
      }

    }

    evals <- evals + (max.i-min.i)

    X1[i+1,] <- curr.x1
    X2[i+1,] <- curr.x2
  }


  # Return state at each iteration, average rejection rate, and average
  # number of slow evaluations per group.

  list (X1=X1, X2=X2, rej=rej/(iters*rep), evals=evals/iters)
}
