# STANDARD METROPOLIS METHOD FOR SAMPLING WITH FAST AND SLOW VARIABLES.
#
# Radford M. Neal, 2010.


# STANDARD METROPOLIS METHOD.  Samples for (x1, x2) using groups of
# Metropolis update.  The first update of each group proposes a change
# to x1 and x2 simultaneously (unless update.1 is TRUE); the remainder 
# of the updates in a group propose changes to x2 only (so that log 
# density evaluations are "fast" for these updates).
#
# Arguments are as follows:
#
#   x1        initial values for the slow variables 
#   x2        initial values for the fast variables 
#   s1        standard deviations for the normal proposals for x1
#   s2        standard deviations for the normal proposals for x2
#   iters     the number of groups of Metropolis updates to do
#   rep       the number of updates in each group 
#   lpr       function that computes the log probability density for the 
#             joint distribution of x1 and x2
#
#   cor.x1,   if these are present (default if absent, both NULL), they
#   cor.x2    must be vectors of positive integers of the same length.
#             When a change is proposed to x2 at an index in cor.x2, the
#             proposal for the element of x1 indexed by the corresponding 
#             value in cor.x1 is modified by subtracting the proposed change 
#             to x2 from the proposed new value for that element of x1.  
#             This applies only to the initial update of both x1 and x2.
#
#   update.1  if TRUE (default is FALSE), update elements one at a time.
#             The total number of updates is therefore greater, and 
#             separate rejection rates are maintained for each element.
#             If cor.x1 and cor.x2 are not NULL, update.1 must be FALSE.
#
# The s1 and s2 arguments may be scalars (same standard deviation for
# all components) or vectors the same lengths as x1 and x2.  
#
# The cor.x1 and cor.x2 arguments are included to allow GP models to use 
# uncorrelated proposals for eta and sigma when the state contains eta and 
# sigma-eta.
#
# The lpr function takes a vector value for the slow variables, x1,
# and a vector value for the fast variables, x2 and returns the log
# probability density (plus any constant) for (x1,x2).  Preferrably,
# the value returned should have attributes that save intermediate
# results of this computation.  When the grid Metropolis function
# calls the log density function for the same value of x1 as
# previously, it passes the value returned then (with its attributes)
# as the third argument.  This allows this subsequent call to be done
# more quickly.  The third argument will be NULL if the result of a
# previous call with the same value for x1 is not available.
#
# The value returned is a list with components x1 and x2, containing
# the iters+1 values from the chain (including the initial value),
# components rej and rej2 containing the rejection rate(s) for the
# Metropolis updates to (x1,x2) and to x2 alone.

std_metropolis <- function (x1, x2, s1, s2, iters, rep, lpr, 
                            cor.x1=NULL, cor.x2=NULL, update.1=FALSE)
{
  if (!is.null(cor.x1) && update.1)
  { stop("Can't have correlated one-element updates")
  }

  # Create matrices to store simulation results, and put in initial state.

  X1 <- matrix(0,iters+1,length(x1))
  X1[1,] <- x1
  X2 <- matrix(0,iters+1,length(x2))
  X2[1,] <- x2

  # Set current state to initial state.

  curr.x1 <- x1
  curr.x2 <- x2

  curr.lpr <- lpr(curr.x1,curr.x2,NULL)

  # Initialize for loop.

  if (update.1)
  { rej <- rep(0,length(x1)+length(x2))
    rej2 <- rep(0,length(x2))
    s1 <- rep(s1,length=length(x1))
    s2 <- rep(s2,length=length(x2))
  }
  else
  { rej <- 0
    rej2 <- 0
  }

  # Loop that does iters groups of Metropolis updates.

  for (i in seq_len(iters))
  { 
    # First update, proposing changes to both x1 and x2.

    if (update.1)
    {
      # Update x1 and x2 one variable at a time.

      for (e in seq_len(length(x1)))
      { 
        prop.x1 <- curr.x1
        prop.x1[e] <- rnorm(1,prop.x1[e],s1[e])

        prop.lpr <- lpr(prop.x1,curr.x2,NULL)

        if (runif(1) < exp (prop.lpr-curr.lpr))
        { curr.x1 <- prop.x1
          curr.lpr <- prop.lpr
        }
        else
        { rej[e] <- rej[e] + 1
        }
      }

      for (e in seq_len(length(x2)))
      { 
        prop.x2 <- curr.x2
        prop.x2[e] <- rnorm(1,prop.x2[e],s2[e])

        prop.lpr <- lpr(curr.x1,prop.x2,curr.lpr)

        if (runif(1) < exp (prop.lpr-curr.lpr))
        { curr.x2 <- prop.x2
          curr.lpr <- prop.lpr
        }
        else
        { rej[length(x1)+e] <- rej[length(x1)+e] + 1
        }
      }
    }
    else # not update.1
    { 
      # Update x1 and x2 simultaneously.

      prop.x1 <- rnorm(length(x1),curr.x1,s1)
      change.x2 <- rnorm(length(x2),0,s2)
      prop.x2 <- curr.x2 + change.x2
      if (!is.null(cor.x1))
      { prop.x1[cor.x1] <- prop.x1[cor.x1] - change.x2[cor.x2]
      }

      prop.lpr <- lpr(prop.x1,prop.x2,NULL)

      if (runif(1) < exp (prop.lpr-curr.lpr))
      { curr.x1 <- prop.x1
        curr.x2 <- prop.x2
        curr.lpr <- prop.lpr
      }
      else
      { rej <- rej + 1
      }
    }

    # Remaining updates, proposing changes to x2 only.

    if (rep>1)
    {
      for (j in 2:rep)
      { 
        if (update.1)
        {
          # Update x2 one variable at a time.

          for (e in seq_len(length(x2)))
          { 
            prop.x2 <- curr.x2
            prop.x2[e] <- rnorm(1,prop.x2[e],s2[e])

            prop.lpr <- lpr(curr.x1,prop.x2,curr.lpr)

            if (runif(1) < exp (prop.lpr-curr.lpr))
            { curr.x2 <- prop.x2
              curr.lpr <- prop.lpr
            }
            else
            { rej2[e] <- rej2[e] + 1
            }
          }
        }
        else # not update.1
        {
          # Update all of x2 simultaneously.

          prop.x2 <- rnorm(length(x2),curr.x2,s2)

          prop.lpr <- lpr(curr.x1,prop.x2,curr.lpr)
  
          if (runif(1) < exp (prop.lpr-curr.lpr))
          { curr.x2 <- prop.x2
            curr.lpr <- prop.lpr
          }
          else
          { rej2 <- rej2 + 1
          }
        }
      }
    }

    X1[i+1,] <- curr.x1
    X2[i+1,] <- curr.x2
  }

  # Return state at each iteration, average rejection rate, and average
  # number of slow evaluations per group.

  list (X1=X1, X2=X2, rej=rej/iters, rej2=rej2/(iters*(rep-1)))
}
